//--------------------------------------------------------------
// File     : main.c
// Datum    : 28.02.2016
// Version  : 1.0
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F746
// Board    : STM32F746-Discovery-Board
// IDE      : OpenSTM32
// GCC      : 4.9 2015q2
// Module   : CubeHAL
// Funktion : Hauptprogramm
//--------------------------------------------------------------


#include "stm32_ub_system.h"
#include "stm32_ub_lcd_480x272.h"
#include "stm32_ub_graphic2d.h"
#include "stm32_ub_uart.h"
#include "stm32f7_audio.h"

// add global define : ARM_MATH_CM7
// add global define : __FPU_PRESENT=1


#include "arm_math.h"
#include "arm_const_structs.h"


#define FFT_DATA                    512  // n*(real+imaginary)  DON'T CHANGE !!
#define FFT_SIZE                FFT_DATA / 2
float32_t FFT_Input[FFT_DATA];
float32_t FFT_Output[FFT_SIZE];
#define FFT_LCD_DATA            FFT_SIZE / 2
uint16_t FFT_LCD[FFT_LCD_DATA];


#define AUDIO_BUFFER_SIZE  1024*2  // size in 16bit words (n * left+right)
uint16_t audio_buffer[AUDIO_BUFFER_SIZE];


#define AUDIO_WAVE_HEIGHT  130
#define FFT_WAVE_HEIGHT    200



arm_cfft_radix4_instance_f32 S;


void init_hardware(void)
{
  UB_LCD_Init();
  UB_LCD_LayerInit_Fullscreen();
  UB_LCD_SetLayer_2();
  UB_LCD_FillLayer(RGB_COL_BLUE);
  UB_LCD_Refresh();

  UB_Uart_Init();
  UB_Uart_SendString(COM1,"Frq-Spectrum",CRLF);
}

void init_audio(uint8_t mode)
{
  if(mode==0) {
    // input = mic, output = ear
    UB_AUDIO_IN_OUT_Init(INPUT_DEVICE_DIGITAL_MICROPHONE_2,OUTPUT_DEVICE_HEADPHONE,90,I2S_AUDIOFREQ_44K);
    UB_AUDIO_IN_Record_Array(audio_buffer, AUDIO_BUFFER_SIZE);
    UB_AUDIO_OUT_Play_Array(audio_buffer, AUDIO_BUFFER_SIZE*2);
  }
  else {
    // input = line, output = none
	UB_AUDIO_IN_Init(INPUT_DEVICE_INPUT_LINE_1,90,I2S_AUDIOFREQ_44K);
	UB_AUDIO_IN_Record_Array(audio_buffer, AUDIO_BUFFER_SIZE);
  }
}

void init_fft(void)
{
	arm_cfft_radix4_init_f32(&S, FFT_SIZE, 0, 1);
}

void fill_fft(void)
{
	uint32_t n;
	int16_t wert;
	int16_t min=audio_buffer[0];
	int16_t max=audio_buffer[0];
	int16_t delta;
	float faktor;

	// seach min and max audio value
	// to range fft_input from +2.0 to -2.0
	for(n=0;n<AUDIO_BUFFER_SIZE;n+=2) {
		wert=audio_buffer[n];
		if(wert>max) max=wert;
		if(wert<min) min=wert;
	}
	delta=max-min;
	faktor=delta/4;

	// convert audio-data in real+imaginary
	for(n=0;n<FFT_DATA;n+=2) {
		wert=audio_buffer[n];
		FFT_Input[n]=(float32_t)(wert/faktor); // real (+2.0 ... -2.0)
		FFT_Input[n+1]=0.0; // imaginary alway zero
	}
}

void calc_fft(unsigned char mode)
{
	uint32_t maxIndex,n;
	float32_t maxValue;
	float32_t wert;

	// calculate fft (FFT_Input --> FFT_Output)
	arm_cfft_radix4_f32(&S, FFT_Input);
	arm_cmplx_mag_f32(FFT_Input, FFT_Output, FFT_SIZE);
	if(mode==0) {
	  // set variable limit
	  arm_max_f32(FFT_Output, FFT_SIZE, &maxValue, &maxIndex);
	}
	else {
	  // set fix limit
	  maxValue=100.0;
	}

	// calculate lcd values
	for(n=0;n<FFT_LCD_DATA;n++) {
		wert=FFT_Output[n];
		FFT_LCD[n]=(uint16_t)(wert/maxValue*(float)(FFT_WAVE_HEIGHT));
	}
}



void draw_fft(void)
{
	uint32_t n;
	uint16_t x,l;

	UB_Graphic2D_DrawStraightDMA(0,LCD_MAXY-1,FFT_SIZE,LCD_DIR_HORIZONTAL,RGB_COL_RED);

	// draw data on screen
	x=0;
	for(n=0;n<FFT_LCD_DATA;n++) {
		l=FFT_LCD[n];
		// draw two lines for each data value
		UB_Graphic2D_DrawStraightDMA(x,LCD_MAXY-l,l,LCD_DIR_VERTICAL,RGB_COL_RED);
		UB_Graphic2D_DrawStraightDMA(x+1,LCD_MAXY-l,l,LCD_DIR_VERTICAL,RGB_COL_RED);
		x+=2;
	}
}

void draw_audio(void)
{
	uint32_t n;
	int16_t wert,delta;
	int16_t min;
	float faktor;
	uint16_t x1,y1,y2,start=0;
	uint8_t status=0;

	delta=5000;
	min=-2500;
	faktor=(float)(AUDIO_WAVE_HEIGHT)/delta;

	// search trigger start (rising edge)
	for(n=0;n<(AUDIO_BUFFER_SIZE-(LCD_MAXX*2));n+=2) {
		wert=audio_buffer[n];
		if(status==0) {
			if(wert<(-100)) status=1;
		}
		else if(status==1) {
			if(wert>=0) {
				start=n;
				break;
			}
		}
	}

	// draw data on screen
	x1=0;
	y1=AUDIO_WAVE_HEIGHT/2;
	for(n=0;n<(LCD_MAXX*2);n+=2) {
		wert=audio_buffer[n+start];
		y2=((wert-min)*faktor);

		UB_Graphic2D_DrawLineNormal(x1,y1,x1+1,y2,RGB_COL_WHITE);
		y1=y2;
		x1++;
	}
}


// called by audio dma interrupt
void update_lcd(void)
{
	static uint16_t display_delay=0;

	display_delay++;
	if(display_delay>2) {
		display_delay=0;
		fill_fft();
		calc_fft(0); // fft-height : 0=fix, 1=variable
		UB_Graphic2D_ClearSreenDMA(RGB_COL_BLUE);
		draw_audio();
		draw_fft();
		UB_LCD_Refresh();
	}
}

int main(void)
{
  // init vom System
  UB_System_Init();


  init_hardware();
  init_audio(0); // input : 0=mic, 1=line
  init_fft();


  while(1) {
	  // nothing to do
  }
}


